"""
Evaluation of parameter sets, i.e. goodness of fit.

Copyright (C) 2020 Daniel Philippus
Full copyright notice located in main.py.
"""

import HydroErr as he
import scipy.stats as sp

def pbias(sim, obs):
    length = len(sim) if len(sim) <= len(obs) else len(obs)
    return 100 * sum([sim[i] - obs[i] for i in range(length)]) / sum(obs[:length])

# Just to have a list of potentially useful tests
# Values are functions taking (simulated, observed) lists (this convention taken from HydroErr) and returning
# the statistic
tests = {
    "r2": he.r_squared,
    "pbias": pbias,
    "rmse": he.rmse,
    "ks_pval": lambda sim, obs: sp.ks_2samp(sim, obs)[1],  # p-value for the null hypothesis (higher is better)
    "ks_stat": lambda sim, obs: sp.ks_2samp(sim, obs)[0],  # ks statistic (smaller is better)
    "paired": lambda sim, obs: sp.ttest_rel(sim, obs)[1],  # p-value (higher is better)
    "mae": he.mae,
    "nse": he.nse
}

# Functions to adjust the above test results so that smaller is better
minimizers = {
    "r2": lambda r2: -r2,
    "pbias": lambda pb: abs(pb),
    "rmse": lambda rmse: rmse,
    "ks_pval": lambda p: -p,
    "ks_stat": lambda ks: ks,
    "paired": lambda p: -p,
    "mae": lambda mae: mae,
    "nse": lambda nse: -nse
}

def minimized(metrics):
    """
    Adjust the given metrics so that smaller is better.
    :param metrics: dictionary of {test: result} from tests
    """
    return {
        test: minimizers[test](metrics[test]) for test in metrics
    }

def fullEval(sim, obs):
    """
    Return all comparison stats
    """
    return {
        test: tests[test](sim, obs) for test in tests
    }

def evaluator(obs, correctDatum, useTests = None):
    """
    Return a function which will return either all (if tests is None) or selected comparison
    stats for a set of simulated values, given the set of observed values.
    :param useTests: list of strings (test names) or None
    :param correctDatum: whether to adjust the datum between obs and sim
    """

    def adjustDatum(sim):
        if not correctDatum:
            return sim
        obs_s = sorted(obs)
        sim_s = sorted(sim)
        count = len(obs) // 20 + 1  # Bottom 5%, +1 in case len(obs) < 20
        adj = (sum(obs_s[:count]) - sum(sim_s[:count])) / count  # Average difference
        return [s + adj for s in sim]

    if useTests is None:
        return lambda sim: fullEval(adjustDatum(sim), obs)
    else:
        return lambda sim: {test: tests[test](adjustDatum(sim), obs) for test in useTests}

def nonDominated(points):
    """
    Return the entries which are not dominated, i.e. there is no entry which is better than them in every metric.
    For NSGA-II, this is part of the algorithm which will presumably be provided by a library.  However, it may
    also be useful to look at non-dominated solutions in semi-manual mode, in which case it is necessary to provide
    a non-domination function.
    :param points: a list of tuples of (value, metrics), metrics should be a list in the same order for all.
        Metrics must all be "lower is better"
    :return: the same format as points, but only those which are non-dominated
    """

    nondom = []
    metrics = [point[1] for point in points]

    for (ix, metric) in enumerate(metrics):
        dominated = False
        for (jx, metric2) in enumerate(metrics):
            if jx != ix:  # Not the same point
                # Dominated if there are no values for which metric is not greater than metric 2
                dominated = False not in [metric[zx] > metric2[zx] for zx in range(len(metric))]
                if dominated:
                    break
        if not dominated:
            nondom.append(points[ix])
    return nondom

def best(points):
    """
    Return the entry which is the best (has the lowest metric).
    :param points: A list of tuples of (value, metric).  Metric must be lower is better
    :return: The best point
    """
    best = None if len(points) == 0 else points[0]
    for point in points:
        if point[1] < best[1]:
            best = point
    return best

def bestN(points, n):
    """
    Return the best n entries (lowest metric), sorted from best to worst
    :param points: A list of tuples of (value, metric).  Metric must be lower is better
    :param n: How many points to return
    :return: A list of the best n points
    """
    if (n == 1):
        return [best(points)]
    else:
        n = n if n <= len(points) else len(points)
        return sorted(points, key = lambda p: p[1])[:n]

def evaluate(obs, sims, correctDatum, metrics = None, useBest = None, usePareto = True, n = 10):
    """
    Evaluate the simulations against obs using the given tests (all if None).  If useBest is specified, get the best
    n based on whatever useBest is.  If usePareto and there are multiple metrics, get the non-dominated results.  If
    usePareto is True and useBest is specified, and there are multiple metrics, then the non-dominated results will
    be passed into the best evaluation.  If there is more than one metric, then either usePareto must be True or useBest
    must be specified.  If there is only one metric, then both usePareto and useBest are ignored and the best n results
    based on the one metric are returned.  Note that the metrics returned are the minimized versions.
    :param obs: observed list
    :param sims: list of (parameters, simulated result).  Simulated result must correspond to obs
    :param metrics: list of metric names to use as specified in the tests dictionary, or None to use all of them
    :param useBest: which metric to sort by, or None not to sort further if usePareto is True
    :param usePareto: whether to only return non-dominated results (including for sorting, if useBest is specified)
    :param n: how many results to return if useBest is specified or if there is only one metric
    :return: list of (parameters, metrics, sim), sorted if useBest is specified, or just one (parameters, metrics)
        if n == 1
    """
    evtr = evaluator(obs, correctDatum, metrics)
    evalf = lambda sim: (sim[0], minimized(evtr(sim[1])), sim[1])
    evaled = [evalf(sim) for sim in sims]
    if (metrics is not None) and (len(metrics) == 1):
        working = bestN([(pt[0], pt[1][metrics[0]], pt[2]) for pt in evaled], n)
        return [(pt[0], evtr(pt[2]), pt[2]) for pt in working]
    else:
        keys = metrics if metrics is not None else list(tests.keys())  # So that the metrics will be in the same order
        working = [(pt[0], [pt[1][key] for key in keys], pt[2]) for pt in evaled]
        if usePareto:
            working = nonDominated(working)
        if useBest is not None:
            keyx = keys.index(useBest)
            working = [(pt[0], pt[1][keyx], pt[2]) for pt in working]  # Only use the one metric
            working = bestN(working, n)
        return [(pt[0], evtr(pt[2]), pt[2]) for pt in working]


if __name__ == "__main__":
    obs = [1,2,4,8,16]
    sims = [
        (1, [1,2,3,4,5]),
        (2, [1,3,6,7,20]),
        (3, [16, 8, 4, 2, 1])
    ]
    print(evaluate(obs, sims))
    print(evaluate(obs, sims, True, useBest = "pbias", n = 1))
    print(evaluate(obs, sims, True, metrics = ["r2", "pbias"]))





"""
Run raspy-cal interactively.

Copyright (C) 2020-2022 Daniel Philippus

This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


from raspy_cal.frontend.input import configSpecify, run
from raspy_cal.frontend import gui
from raspy_cal.settings import Settings
from sys import argv

msg = """Raspy-Cal interactive command-line interface.

Copyright (C) 2022 Daniel Philippus
This program comes with ABSOLUTELY NO WARRANTY.
    This is free software, and you are welcome to redistribute it
    under certain conditions.    For details see the LICENSE file
    at https://github.com/quantum-dan/raspy-cal.

Note: in order to run somewhat more economically, you can use:
python raspy_cal/main.py <project path> <stage file path> <output file path>.
Alternatively, to use a config file, run: `python main.py <config file path>`
or `raspy-cal.exe <config file path>`.
"""

"""
Testing with LAR data:
* F37B = Compton Creek : CC : 23350
* F45B = Rio Hondo Chnl : RHC : 7000
* F300 = Upper LAR : Above RH : 195289.1
* F319 = LA River : Below CC : 21500
"""

locs = {
    "F37B": {"river": "Compton Creek", "reach": "CC", "rs": "23350."},
    "F45B": {"river": "Rio Hondo Chnl", "reach": "RHC", "rs": "7000"},
    "F300": {"river": "Upper LAR", "reach": "Above RH", "rs": "195289.1"},
    "F319": {"river": "LA River", "reach": "Below CC", "rs": "21500"}
}

basepath = "V:\\LosAngelesProjectsData\\HEC-RAS\\raspy_cal\\"  # for testing

def run():
    settings = Settings()
    if len(argv) == 4:
        settings.specify(project=argv[1], stagef=argv[2], outf=argv[3])
        settings.interactive()
        run(settings)
    # elif len(argv) == 2:  # for testing
    #     if argv[1] == "LAR":  # Test with LA project
    #         gage = input("Gage (F37B, F45B, F300, F319): ")
    #         for metric in ["r2", "pbias", "rmse", "ks_pval", "ks_stat", "paired", "mae", "nse"]:
    #             print("Running %s with %s" % (gage, metric))
    #             specify(
    #                 project=basepath + "LAR\\FullModel.prj",
    #                 stagef=basepath + "data\\" + gage + ".csv",
    #                 river=locs[gage]["river"],
    #                 reach=locs[gage]["reach"],
    #                 rs=locs[gage]["rs"],
    #                 outf=basepath + "data\\Out_" + gage + "_" + metric + ".txt",
    #                 plot=True,
    #                 auto=True,
    #                 metrics=[metric],
    #                 nct=10,
    #                 evals=50,
    #                 fileN="05",
    #                 slope=0.001
    #             )
    #     elif argv[1] == "CMDTEST":  # Test the text interface
    #         basepath = "V:\\LosAngelesProjectsData\\HEC-RAS\\"
    #         specify(
    #             project=basepath + "raspy\\DemoProject\\project.prj",
    #             stagef=basepath + "raspy_cal\\DemoStage.csv",
    #             outf=basepath + "raspy_cal\\DemoOut.txt"
    #         )
    #     elif argv[1] == "CMD":
    #         print(msg)
    #         specify()
    #     elif argv[1] == "GUI":  # Test the GUI (when implemented)
    #         gui.main()
    #     else:
    #         configSpecify(argv[1], run=True)
    elif len(argv) == 2:
        if argv[1] == "CMD":
            settings.interactive()
            run(settings)
        elif argv[1].lower() in ["h", "-h", "help", "--help"]:
            print(msg)
        else:
            settings = configSpecify(argv[1], settings)
            run(settings)
    else:
        print("Run python main.py <config file path> or raspy-cal.exe <config file path> to load a config file \
    in the command line version. \
    Run python main.py CMD or raspy-cal.exe CMD to use the command line version.")
        gui.main(settings)


if __name__ == "__main__":
    run()


"""
Accept relevant inputs and put them into a useful format.

Copyright (C) 2020-2022 Daniel Philippus
Full copyright notice located in main.py.
"""

from raspy_cal.default import Model
from raspy_cal.lowlevel import runSims
from raspy_cal.midlevel.eval import evaluate, minimized, fullEval, tests, evaluator
from raspy_cal.midlevel.params import paramSpec, genParams
from raspy_cal.frontend.display import evalTable, compareAllRatingCurves, nDisplay
from raspy_cal.midlevel.data import getUSGSData, prepareUSGSData, singleStageFile
from raspy_cal.midlevel.calibrators import nstageIteration, nstageSingleRun
from raspy_cal.settings import Settings

from platypus import NSGAII, Problem, Real, nondominated # https://platypus.readthedocs.io/en/latest/getting-started.html#defining-constrained-problems
from urllib.request import urlopen


def parseConfigText(text, parsers):
    """
    Parse the text of a config file.  Format is Keyword: Value one per line.  Lines starting with # are
    ignored.
    :param text: the text of the config file.
    :param parsers: a dictionary of {keyword: function}, where function takes the string and returns
        the relevant value.  Lowercase keywords.
    :return: a dictionary of {keyword: value}
    """
    lines = [(line, line.index(":")) for line in text.split("\n") if not line.startswith("#")
             and not line == ""]
    # Before colon : after colon
    stringvals = { line[0][:line[1]].strip().lower() : line[0][line[1]+1:].strip() for line in lines }
    # So that values will be None if they aren't present in the file
    result = {k: None for k in parsers}
    for v in stringvals:
        if stringvals[v] != "":
            result[v] = parsers[v](stringvals[v])
    return result

def configSpecify(confPath, settings = Settings(), run = True):
    """
    Parse all of the arguments for specify and run it.
    :param confPath: path to the config file, or None to return example config file format
    :return: config values or example file format
    """
    id = lambda x: x
    toBool = lambda x: x in ["True", "true", "t", "T"]
    parsers = {
        "version": id,
        "project": id,
        "stagef": id,
        "river": id,
        "reach": id,
        "rs": id,
        "nct": int,
        "outf": id,
        "plot": toBool,
        "auto": toBool,
        "evals": int,
        "metrics": lambda x: [i.strip() for i in x.split(",")],  # format: r2,rmse,pbias
        "filen": id,
        "slope": float,
        "usgs": id,
        "flowcount": int,
        "enddate": id,
        "startdate": id,
        "period": int,
        "si": toBool,
        "datum": toBool
    }
    if confPath is not None:
        with open(confPath) as f:
            data = f.read()
        vals = parseConfigText(data, parsers)
        if run:
            settings.specify(
                project=vals["project"], stagef=vals["stagef"], river=vals["river"], reach=vals["reach"],
                rs=vals["rs"], nct=vals["nct"], outf=vals["outf"], plot=vals["plot"], auto=vals["auto"],
                evals=vals["evals"], metrics=vals["metrics"], fileN=vals["filen"], slope=vals["slope"],
                usgs=vals["usgs"], flowcount=vals["flowcount"], enddate=vals["enddate"], startdate=vals["startdate"],
                period=vals["period"], si=vals["si"], correctDatum=vals["datum"],
                version=vals["version"]
            )
            settings.interactive()
            return settings
        return vals
    else:
        return """
# Example configuration file.  Fill out the values below to use and run with `python main.py <path>` or
# `raspy-cal.exe <path>` to use this config file.  Comment out lines you don't want to specify with `#`
# at the start of the line; information not provided will be requested when running as needed.  Not all
# information is required, e.g. if usgs is specified stagef will be ignored.
project: C:\\PathToHECRASProject\\project.prj
version: 507
stagef: C:\\PathToStagefileIfUsed\\stagefile.csv
river: South Platte River
reach: Above Confluence Park
rs: 12345.0*
nct: 10
outf: C:\\PathToOutputFile\\outfile.csv
plot: True
auto: True
evals: 100
metrics: r2,rmse,pbias
filen: 04
slope: 0.025
usgs: 09429100
flowcount: 100
enddate: 2020-02-26
startdate: 2019-02-28
period: 500
si: False
"""


def run(settings):
    auto = settings.auto
    if auto:
        autoIterate(settings)
    else:
        iterate(settings)


def iterate(settings, model=None, rand=None):
    """
    Iterate over n options until the user narrows it down to a good choice.
    Note that providing an n of 0 will
    cause HEC-RAS to crash.
    :return: final ns
    """
    model = Model(settings.project, settings.version) if model is None else model
    rand = input("Enter Y to use random parameter generation: ") in ["y", "Y"]\
        if rand is None else rand
    plotpath = ".".join(settings.outf.split(".")[:-1]) + ".png"
    cont = True
    while cont:
        nmin = float(input("Enter minimum n: "))
        nmax = float(input("Enter maximum n: "))
        best = nstageIteration(model,
                               settings.river,
                               settings.reach,
                               settings.rs,
                               settings.stage,
                               settings.nct,
                               rand,
                               nmin,
                               nmax,
                               settings.metrics,
                               settings.datum)
        # Show plot (if specified) but don't save anything
        nDisplay(best, settings.flow, settings.stage, None, None,
                 settings.plot, settings.datum, settings.si)
        cont = input("Continue?  Q or q to quit and write results: ")\
            not in ["q", "Q"]
        if not cont:
            # Save the plot and CSV
            nDisplay(best, settings.flow, settings.stage,
                     plotpath, settings.outf, False, settings.datum,
                     settings.si)


def autoIterate(settings, model=None):
    """
    Automatically iterate with NSGA-II
    """
    keys = settings.metrics  # ensure same order
    evalf = evaluator(settings.stage,
                      useTests=keys,
                      correctDatum=settings.datum)
    plotpath = ".".join(settings.outf.split(".")[:-1]) + ".png"
    count = 1
    print("Running automatic calibration")

    def manningEval(vars):
        n = vars[0]
        metrics = minimized(
            nstageSingleRun(model,
                            settings.river,
                            settings.reach,
                            settings.rs,
                            settings.stage,
                            n,
                            keys,
                            settings.datum)
        )
        values = [metrics[key] for key in keys]
        constraints = [-n, n - 1]
        nonlocal count
        print("Completed %d evaluations" % count)
        count += 1
        return values, constraints
    c_type = "<0"
    # 1 decision variable, len(keys) objectives, and 2 constraints
    problem = Problem(1, len(keys), 2)
    problem.types[:] = Real(0.001, 1)  # range of decision variable
    problem.constraints[:] = c_type
    problem.function = manningEval

    algorithm = NSGAII(problem, population_size=settings.nct)
    algorithm.run(settings.evals)
    # nondom: list of Solutions - wanted value is variables[0]
    nondom = nondominated(algorithm.result)
    nondomNs = [sol.variables[0] for sol in nondom]
    results = runSims(model, nondomNs, settings.river,
                      settings.reach, len(settings.stage),
                      range=[settings.rs])
    resultPts = [(nondomNs[ix], [results[ix][settings.rs][jx] for jx in range(
        1, len(settings.stage) + 1)]) for ix in range(len(nondomNs))]
    metrics = [(res[0], evalf(res[1]), res[1]) for res in resultPts]
    nDisplay(metrics, settings.flow, settings.stage, plotpath,
             settings.outf, settings.plot, settings.datum, settings.si)
    return metrics



# -*- coding: utf-8 -*-
"""
Created on Mon May 23 13:20:06 2022

@author: dphilippus
"""

"""
Create a class that holds arguments to pass around.

Copyright (C) 2022 Daniel Philippus
Full copyright notice located in main.py.
"""

from raspy_cal.midlevel.data import (getUSGSData,
                                     prepareUSGSData, singleStageFile)
from raspy_cal.midlevel.eval import tests


class Settings(object):
    # Stores calibrator settings so that there is a consistent format
    # for passing them around instead of massive lists of arguments.
    def __init__(self):
        # Everything is None to begin with (empty object).
        self.version = None
        self.project = None
        self.stagef = None
        self.river = None
        self.reach = None
        self.rs = None
        self.nct = None
        self.outf = None
        self.plot = None
        self.auto = None
        self.evals = None
        self.metrics = None
        self.fileN = None
        self.slope = None
        self.usgs = None
        self.flowcount = None
        self.enddate = None
        self.startdate = None
        self.period = None
        self.datum = None
        self.si = None
        self.flow = None
        self.stage = None

    def specify(self,
                project=None,
                stagef=None,
                river=None,
                reach=None,
                rs=None,
                nct=None,
                outf=None,
                plot=None,
                auto=None,
                evals=None,
                metrics=None,
                fileN=None,
                slope=None,
                usgs=None,
                flowcount=None,
                enddate=None,
                startdate=None,
                period=None,
                correctDatum=None,
                si=False,
                version=None,
                stage=None,
                flow=None
                ):
        # Set up initial settings with one call.

        # Arguments retrieved from frontend.input.specify as of
        # first writing.
        # Only overwrite settings if argument is provided.
        if project is not None:
            self.project = project
        if stagef is not None:
            self.stagef = stagef
        if stage is not None:
            self.stage = stage
        if flow is not None:
            self.flow = flow
        if river is not None:
            self.river = river
        if reach is not None:
            self.reach = reach
        if rs is not None:
            self.rs = rs
        if nct is not None:
            self.nct = nct
        if outf is not None:
            self.outf = outf
        if plot is not None:
            self.plot = plot
        if auto is not None:
            self.auto = auto
        if evals is not None:
            self.evals = evals
        if metrics is not None:
            self.metrics = metrics
        if fileN is not None:
            self.fileN = fileN
        if slope is not None:
            self.slope = slope
        if usgs is not None:
            if self.stagef is None:
                self.usgs = usgs
            else:
                self.usgs = ""
        if flowcount is not None:
            self.flowcount = flowcount
        if enddate is not None:
            self.enddate = enddate
        if startdate is not None:
            self.startdate = startdate
        if period is not None:
            self.period = period
        if correctDatum is not None:
            self.datum = correctDatum
        if si is not None:
            self.si = si
        if version is not None:
            self.version = version

    def interactive(self):
        # Get settings from user via interactive command line usage.
        def getUSGS(usgs, flowcount, enddate, startdate, period, si):
            flowcount = int(input("Approx. how many flows to retrieve: ")
                            ) if flowcount is None else flowcount
            enddate = input(
                "End date or leave blank for today: ") if enddate is None else\
                enddate
            startdate =\
                input("Start date or leave blank for 1 week ago or period: ")\
                if startdate is None and\
                (period is None or period == "")\
                else startdate
            period = input(
                "Period or leave blank for 1 week or start date: ") if period\
                is None else period
            return prepareUSGSData(
                getUSGSData(usgs, enddate, startdate, period, si=si),
                flowcount
            )

        self.version =\
            input("Enter HEC-RAS version without dots (e.g., 507, 631):")\
            if self.version is None else self.version
        self.project = input(
            "Enter project path (including .prj file): ") if self.project is\
            None else self.project
        self.usgs = input(
            "USGS gage number or leave blank to use a stage file: ") if\
            self.usgs is None else self.usgs
        if self.usgs == "":
            self.stagef = input(
                "Enter path to stage file: ") if self.stagef is None else\
                self.stagef
        (self.flow, self.stage) = singleStageFile(self.stagef) if\
            self.usgs == "" else getUSGS(
            self.usgs, self.flowcount, self.enddate, self.startdate,
            self.period)
        self.outf = input(
            "Enter output file path or nothing to not have one: ") if\
            self.outf is None else self.outf
        self.fileN = input(
            "Enter flow file number to write (default 01): ") if self.fileN is\
            None else self.fileN
        self.fileN = "01" if self.fileN == "" else self.fileN
        self.slope = input(
            "Enter slope for normal depth (default 0.001): ") if self.slope is\
            None else self.slope
        self.slope = float(0.001) if self.slope == "" else float(self.slope)
        if self.metrics is None:
            self.metrics = [] if input("Enter Y to specify metrics: ") in [
                "Y", "y"] else None
            keys = list(tests.keys())
            if self.metrics == []:
                inp = ""
                while inp not in ["D", "d"]:
                    print("Available metrics: %s" % keys)
                    print("Selected metrics: %s" % self.metrics)
                    inp = input("Enter a metric to add it or D if done: ")
                    if inp in keys:
                        self.metrics.append(inp)
                    elif inp not in ["D", "d"]:
                        print("Warning: entered metric is not an option.")
            if self.metrics == []:
                self.metrics = None
        self.river = input(
            "River name: ") if self.river is None else self.river
        self.reach = input(
            "Reach name: ") if self.reach is None else self.reach
        self.rs = input("River station: ") if self.rs is None else self.rs
        self.auto =\
            input("Enter Y to use automatic calibration\
 (default: interactive): ") in [
                "Y", "y"] if self.auto is None else self.auto
        self.plot = input("Enter N to not plot results (default: plot): ")\
            not in [
            "N", "n"] if self.plot is None else self.plot
        self.nct = int(input("Number of n to test each iteration: ")
                       ) if self.nct is None else self.nct
        self.evals = int(input("How many evaluations to run? ")
                         ) if self.auto and self.evals is None else self.evals
        self.datum =\
            input("Enter Y to correct datum (default: no correction): ") in [
                "Y", "y"] if self.datum is None else self.datum
        self.si = input("Enter Y if HEC-RAS project and flow data are in SI\
 units (default: US customary): ") in [
            "Y", "y"] if self.si is None else self.si



"""
Generate parameter combinations.  Currently, it is only intended to be used with Manning's n, but
multiple parameters could be used in the future.

Copyright (C) 2020 Daniel Philippus
Full copyright notice located in main.py.
"""

import random

def paramSpec(name, min, max, n = 10, random = False):
    """
    Specifies a parameter suitable for use in genParams.  Just doing it this way because a class
    seemed excessive.
    :param name: name of the parameter (string)
    :param min: minimum value of parameter
    :param max: maximum value of parameter
    :param n: how many variants to generate
    :param random: generate variants randomly instead of evenly spaced
    :return: just a dictionary with the above information
    """
    return {
        "name": name,
        "min": min,
        "max": max,
        "n": n,
        "random": random
    }

def mkCombs(lists):
    """
    Make all possible combinations of the elements in the lists, in order
    :param lists: list of lists
    :return: list of combinations (also a list of lists)
    """
    out = [[]]
    for l in lists:
        out = [ox + [lx] for lx in l for ox in out]
    return out

def genParams(paramSpecs, dicts = True):
    """
    Generate combinations of parameters.
    :param paramSpecs: a list of dictionaries like those generated by paramSpec()
    :param dicts: return a list of dictionaries with the parameters named, instead of just a list of lists
    :return: a list of dictionaries {name: value} of parameter combinations, or a list of lists
    """
    out = []
    plists = []
    for ps in paramSpecs:
        if ps["random"]:
            plists.append([
                random.uniform(ps["min"], ps["max"]) for _ in range(ps["n"])
            ])
        else:
            plists.append(
                [i * (ps["max"] - ps["min"]) / (ps["n"] - 1) + ps["min"] for i in range(ps["n"])]
            )
    combs = mkCombs([range(ps["n"]) for ps in paramSpecs])
    for cx in combs:
        if dicts:
            out.append({
                paramSpecs[i]["name"]: plists[i][cx[i]] for i in range(len(cx))
            })
        else:
            out.append([
                plists[i][cx[i]] for i in range(len(cx))
            ])
    if dicts or len(out[0]) > 1:
        return out
    else:
        return [i[0] for i in out] # Allows for use with a single parameter

if __name__ == "__main__":
    # For debugging
    pspecs1 = [paramSpec("a", 1, 5, 5), paramSpec("r", -5, 0, random = True)]
    pspecs2 = [paramSpec("n", 0.01, 0.1, 10, True)]
    # print(genParams(pspecs1, dicts = False))
    print(genParams(pspecs2, dicts = False))





"""
Functions to display results.

Copyright (C) 2020 Daniel Philippus
Full copyright notice located in main.py.
"""

import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter


def csv(list):
    return "\n".join([",".join(row) for row in list])

def space(entry, width = 9, after = False):
    """
    Add spaces to the entry so that it is the appropriate width.  Add spaces before unless after is true.
    """
    if len(entry) < width:
        if after:
            return entry + " " * (width - len(entry))
        else:
            return " " * (width - len(entry)) + entry
    else:
        return entry

def evalTable(params, metricSets, paramName = "n", string = True):
    """
    Make a table for printing to the console of params vs metrics.
    :param params: list of the param values
    :param metricSets: list of corresponding metric sets (dictionaries of name: value)
    :param paramName: what to call the parameter
    :param string: whether to return it as a string (if not, then a list)
    :return: either a list of lists (inner lists = rows) or the table as a string
    """
    keys = list(metricSets[0].keys())
    header = [space(i) for i in [paramName] + keys] if string else [paramName] + keys
    rows = [header]
    for ix in range(len(params)):
        row = ["%.3f" % params[ix]] + ["%.3f" % metricSets[ix][k] for k in keys]
        if string:
            rows.append([space(i) for i in row])
        else:
            rows.append(row)
    if string:
        return "\n".join([" ".join(row) for row in rows])
    else:
        return rows

def nDisplay(results, flow, obs, plotpath=None, csvpath=None, plot=True, correctDatum = False, si = False):
    """
    Wrapper for displayOutputs using 1-D/Manning's n defaults.
    :param results: [(parameter, metrics, stage)]
    :param flow: flow values
    :param obs: observed stage
    :param plotpath: path to save plot
    :param csvpath: path to save CSV
    :param plot: whether to plot
    :return: list version of result table
    """
    return displayOutputs("n", results, flow, obs, "Rating Curves Comparison", "Flow (cfs)" if not si else "Flow (cms)",
                          "Stage (ft)" if not si else "Stage (m)", True, True, plotpath, plot, csvpath, correctDatum, si)


def displayOutputs(paramName, results, obsX, obsY, title="", xlab="", ylab="", xlog=True, ylog=True, plotpath=None,
                   plot=True, csvpath=None, correctDatum = False, si = False):
    """
    Print metric table, show plot (if specified), and save plot (if specified).
    :param paramName: name of calibration parameter
    :param results: list of [(parameter, metrics, output list)]
    :param obsX: x values (e.g. flow) for plotting
    :param obsY: observed y values
    :param si: 0 for US units, 1 for metric, 2 to convert US units to metric
    :param title: plot title
    :param xlab: plot x label
    :param ylab: plot y label
    :param xlog: log scale for x?
    :param ylog: log scale for y?
    :param plotpath: where to save plot or None not to
    :param csvpath: where to save CSV version of metrics table or None not to
    :param plot: whether to plot
    :param si: SI units
    :return: list version of result table
    """
    (params, metrics, timeseries) = ([res[0] for res in results], [res[1] for res in results],
                                     [(res[0], res[2]) for res in results])
    stringTable = evalTable(params, metrics, paramName, True)
    listTable = evalTable(params, metrics, paramName, False)
    print(stringTable)
    qlab = "Q.cms" if si else "Q.cfs"
    stlab = "ObsStage.m" if si else "ObsStage.ft"
    simlab = "SimStage.m.n=" if si else "SimStage.ft.n="
    if csvpath is not None:
        with open(csvpath, "w") as f:
            f.write(csv(listTable))
        parts = csvpath.split(".")
        datapath = ".".join(parts[:-1]) + "-data.csv"
        datatable = [[qlab, stlab] + [simlab + str(p) for p in params]] + \
                     [[str(obsX[i]), str(obsY[i])] + [str(r[2][i]) for r in results] for i in range(len(obsX))]
        with open(datapath, "w") as f:
            f.write(csv(datatable))
    if (plotpath is not None) or plot:
        compareAllRatingCurves(obsX, obsY, timeseries, plot, plotpath, paramName, title, xlab, ylab, xlog, ylog,
                               correctDatum)
    return listTable


def compareRatingCurve(flows, obs, sim, si=False):
    plt.plot(flows, obs, label = "Observed")
    plt.plot(flows, sim, label = "Simulated")
    plt.xlabel("Flow (cfs)" if not si else "Flow (cms)")
    plt.ylabel("Depth (ft)" if not si else "Flow (cfs)")
    plt.legend()
    plt.show()

# see https://matplotlib.org/3.1.0/api/markers_api.html
markers = [c for c in ".ov^<>1234sP*+xXDd|_"] + ["$%s$" % (chr(c + ord("A"))) for c in range(26)]

def compareAllRatingCurves(x, obs, sims, display=True, path=None, paramName="n",
                           title="Rating Curves Comparison", xlab="Flow (cfs)",
                           ylab="Depth (ft)", xlog=True, ylog=True, correctDatum = False):
    """
    Compare all provided rating curves.  Default arguments are depth vs flow calibrating n.
    :param x: list of x variable
    :param obs: list of observed y variable
    :param sims: list of (parameter, [simulated ys])
    :param display: whether to show the plot
    :param path: where to save the plot, or None not to
    :param paramName: name of parameter (for legend)
    """
    def adjustDatum(sim):
        if not correctDatum:
            return sim
        obs_s = sorted(obs)
        sim_s = sorted(sim)
        count = len(obs) // 20 + 1  # Bottom 5%, +1 in case len(obs) < 20
        adj = (sum(obs_s[:count]) - sum(sim_s[:count])) / count  # Average difference
        return [s + adj for s in sim]

    plt.clf()  # Prevent previous plots being shown on the same axes
    fig = plt.plot(x, obs, label = "Observed")[0]
    for (ix, (par, sim)) in enumerate(sims):
        plt.plot(x, adjustDatum(sim), label = "Simulated (%s = %.3f)" % (paramName, par), marker=markers[ix % len(markers)])
    if xlog:
        plt.xscale("log")
    if ylog:
        plt.yscale("log")
    plt.xlabel(xlab)
    plt.ylabel(ylab)
    if x[-1]/x[0] < 50: # only use minor x-axis labels if there won't be enough major labels
        fig.axes.xaxis.set_minor_formatter(FormatStrFormatter("%.1f"))
    fig.axes.yaxis.set_minor_formatter(FormatStrFormatter("%.2f"))
    plt.title(title)
    plt.legend()
    if path is not None:
        plt.savefig(path)
    if display:
        plt.show()

